/*  $Id: NodeAddress.cpp,v 1.29 2003/05/22 16:43:08 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Implementation of the <code>NodeAddress</code> class.

    @author Holger Wunsch */


/** Although feature structures can contain shared data, the visible
    properties of two instances of a shared feature structure (e.g.
    visible vs. hidden, expanded vs. collapsed) can be different. In order
    to do this, every node that can be displayed must have a unique address
    that not only encodes which data node it is associated with but also its
    exact position in the overall feature structure.
    This unique address is essentially a path from the topmost node in the
    overall structure to the substructure it identifies, very much like
    the paths in a file system.
    The format of a node address is as follows:
    Every single node in the data structure passed to grisu specifies an
    integer number in the "id" field of the record (see the communications
    protocol EBNF). This number is converted into an array of individual
    integers, each representing one indivdual digit, so 2436 becomes
    [2,4,3,6]. The full address of a node is calculated by concatenating
    all the id's of all the nodes from the root to this node.
    The two uppper bits in each entry are reserved, where 0x80 indicates
    that the id of a new node starts at this position, and 0x40 indicates
    that the id of a node ends at this position. */

#include <string.h>
#include "NodeAddress.h"

NodeAddress::NodeAddress() {
  addr = NULL;
  len = 0;
}

NodeAddress::NodeAddress(NodeAddress& na) {
  len = na.len;
  addr = new unsigned char[len];

  memcpy(addr, na.addr, len);
}

NodeAddress::~NodeAddress() {
  if (addr)
    delete[] addr;
}

NodeAddress::NodeAddress(const NodeAddress& addrParent, GenericNode* nd) {
  unsigned int idlen = strlen(nd->id);
  len = idlen + addrParent.len;
  addr = new unsigned char[len];
  int i;

  for (i = 0; i < idlen; i++) {
    addr[i] = nd->id[i] - '0';
    if (i == 0)
      addr[i] |= 0x80;
    if (i == idlen - 1)
      addr[i] |= 0x40;
  }
  memcpy(&(addr[i]), addrParent.addr, addrParent.len);
}

NodeAddress::NodeAddress(const NodeAddress& addrParent, char* id) {
  unsigned int idlen = strlen(id);
  len = idlen + addrParent.len;
  addr = new unsigned char[len];
  int i;

  for (i = 0; i < idlen; i++) {
    addr[i] = id[i] - '0';
    if (i == 0)
      addr[i] |= 0x80;
    if (i == idlen - 1)
      addr[i] |= 0x40;
  }
  memcpy(&(addr[i]), addrParent.addr, addrParent.len);
}


bool NodeAddress::contains(char* idc) const {
  int idlen = strlen(idc);
  unsigned char id[idlen];
  bool cont = false;

  for (int i = 0; i < idlen; i++) {
    id[i] = idc[i] - '0';
    if (i == 0)
      id[i] |= 0x80;
    if (i == idlen - 1)
      id[i] |= 0x40;
  }

  for (int i = 0; i < len; i++) {
    if (addr[i] == id[0]) {
      cont = true;
      for (int j = 0; j < idlen; j++)
	if (addr[i + j] != id[j]) {
	  cont = false;
	  break;
	}
      if (cont)
	return true;
    }
  }
  return false;
}

bool NodeAddress::operator == (const NodeAddress& na) const {
  if (len == 0) {
    if (na.len == 0)
      return true;
    else
      return false;
  }
 
  if (len == na.len)
    return (memcmp(addr, na.addr, len) == 0);
  else
    return false;
}

NodeAddress NodeAddress::operator = (const NodeAddress& na) {
  if (na.len == 0) {
    if (addr)
      delete [] addr;
    addr = NULL;
    len = 0;
    return *this;
  }

  len = na.len;
  addr = new unsigned char[len];

  memcpy(addr, na.addr, len);

  return *this;
}


void NodeAddress::parentAddress(NodeAddress& pa) {
  int i;
  
  if (pa.len > 0) {
    delete [] pa.addr;
    pa.len = 0;
  }

  for (i = 0; i < len; i++)
    if (addr[i] & 0x40)
      break;

  if (++i < len) {
    pa.len = len - i;
    pa.addr = new unsigned char[pa.len];
    memcpy(pa.addr, &(addr[i]), pa.len);
  }
}


/** Returns the node ID of the node this address points to in a static buffer.
    @return a pointer to a static <code>char</code> buffer that contains the node ID. */
const char* NodeAddress::nodeID(void) const {
  static char id[20];
  int index;

  for (index = 0; index < len; index++) {
    id[index] = (addr[index] & 0x0F) + '0';
    if (addr[index] & 0x40)
      break;
  }

  id[++index] = '\x0';

  return (const char*) id;
}
  
  
void NodeAddress::print() const {
  for (unsigned int i = 0; i < len; i++) {
    if (addr[i] & 0x80)
      printf("(");
    printf("%d", addr[i] & 0x0F);
    if (addr[i] & 0x40)
      printf(")");
  }
  printf("\n");
}

void NodeAddress::sprint(char* buf) {
  int bptr = 0;
  char nbuf[50];
  for (unsigned int i = 0; i < len; i++) {
    if (addr[i] & 0x80)
      buf[bptr++] = '(';
    sprintf(nbuf, "%d", addr[i] & 0x0F);
    memcpy(&buf[bptr], nbuf, strlen(nbuf));
    bptr += strlen(nbuf);
    if (addr[i] & 0x40)
      buf[bptr++] = ')';
  }
  buf[bptr] = 0x00;
}  
