/*  $Id: canvas.cpp,v 1.32 2003/05/22 16:43:09 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    Implementation of the <code>Canvas</code>, with exception of all
    drawing routines, which can be found in drawer.cpp.

    @author Holger Wunsch */

#include <qaccel.h>
#include <qdir.h>
#include <qregexp.h>
#include <qstringlist.h>

#include "global.h"
#include "canvas.h"
#include "node.h"

#include "canvas.moc"

Canvas::Canvas(Options *op, MyScrollBar *vscroll, MyScrollBar *hscroll,
	       Document *doc, GenericNode* rootNd, QWidget *parent, const char *name )
  :QWidget(parent, name)
{
  QAccel* accel = new QAccel(this);


  options=op;
  

  setPalette(options->backgroundColor);

  document=doc;

  if (rootNd == NULL) {
    rootNode = document->data->at(0);
  }
  else
    rootNode = rootNd;

  toKlick.setAutoDelete(true);

  resetToClick();
  
  maxSize=QSize(100,100);
  v_scroll=vscroll;
  h_scroll=hscroll;
  offsetX=offsetY=0;

  connect( v_scroll, SIGNAL(valueChanged(int)),
	   this, SLOT(scrollingSlot(int)) );
  connect( h_scroll, SIGNAL(valueChanged(int)),
	   this, SLOT(scrollingSlot(int)) );

  accel->connectItem(accel->insertItem(Key_Down),
		     this,
		     SLOT(scrollLineDownSlot()));
  accel->connectItem(accel->insertItem(Key_Up),
		     this,
		     SLOT(scrollLineUpSlot()));
  accel->connectItem(accel->insertItem(Key_Left),
		     this,
		     SLOT(scrollLineLeftSlot()));
  accel->connectItem(accel->insertItem(Key_Right),
		     this,
		     SLOT(scrollLineRightSlot()));
  accel->connectItem(accel->insertItem(CTRL + Key_Down),
		     this,
		     SLOT(scrollPageDownSlot()));
  accel->connectItem(accel->insertItem(Key_PageDown),
		     this,
		     SLOT(scrollPageDownSlot()));
  accel->connectItem(accel->insertItem(CTRL + Key_Up),
		     this,
		     SLOT(scrollPageUpSlot()));
  accel->connectItem(accel->insertItem(Key_PageUp),
		     this,
		     SLOT(scrollPageUpSlot()));
  accel->connectItem(accel->insertItem(CTRL + Key_Left),
		     this,
		     SLOT(scrollPageLeftSlot()));
  accel->connectItem(accel->insertItem(CTRL + Key_Right),
		     this,
		     SLOT(scrollPageRightSlot()));


  dragging = false;
  setMouseTracking(true);

  highlight = NULL;
  foundMark = NULL;
  
  dispNodes = new IdTree(document->maxID() + 1);
}


Canvas::~Canvas() {
  delete dispNodes;
}


void Canvas::scrollingSlot(int ) {
  offsetX=h_scroll->value();
  offsetY=v_scroll->value();

  repaint(false);

  ((ViewCanvas*) parentWidget()->parentWidget())->setStatusBarMessage("");
}


void Canvas::scrollLineDownSlot(void) {
  v_scroll->setValue(v_scroll->value() + v_scroll->lineStep());
}


void Canvas::scrollLineUpSlot(void) {
  v_scroll->setValue(v_scroll->value() - v_scroll->lineStep());
}


void Canvas::scrollLineRightSlot(void) {
  h_scroll->setValue(h_scroll->value() + h_scroll->lineStep());
}


void Canvas::scrollLineLeftSlot(void) {
  h_scroll->setValue(h_scroll->value() - h_scroll->lineStep());
}


void Canvas::scrollPageDownSlot(void) {
  v_scroll->setValue(v_scroll->value() + v_scroll->pageStep());
}


void Canvas::scrollPageUpSlot(void) {
  v_scroll->setValue(v_scroll->value() - v_scroll->pageStep());
}


void Canvas::scrollPageRightSlot(void) {
  h_scroll->setValue(h_scroll->value() + h_scroll->pageStep());
}


void Canvas::scrollPageLeftSlot(void) {
  h_scroll->setValue(h_scroll->value() - h_scroll->pageStep());
}


/** Called when a wheel event occurs (wheel mice only).
    @param we the wheel event class. */
void Canvas::wheelEvent(QWheelEvent* we) {
  if (we->state() & ControlButton) {
    if (we->delta() > 0)
      h_scroll->setValue(h_scroll->value() - 2 * h_scroll->lineStep());
    else
      h_scroll->setValue(h_scroll->value() + 2 * h_scroll->lineStep());
  }
  else {
    if (we->delta() > 0)
      v_scroll->setValue(v_scroll->value() - 2 * v_scroll->lineStep());
    else
      v_scroll->setValue(v_scroll->value() + 2 * v_scroll->lineStep());
  }
    
  we->accept();
}


void Canvas::mousePressEvent(QMouseEvent *mo) {
  DisplayableNode* dn;

  if (mo->button() == RightButton) {
    klickedOn = whereKlicked(mo->pos());
    if (klickedOn) {
      popup = new KPopupMenu(this);
      switch (klickedOn->type) {
      case NODETYPE_TYPE_NODE: {
	char itemText[500];
	NodeAddress addrParent;
	DisplayableNode* dnd;
	klickedOn->addr.parentAddress(addrParent);

	dnd = dispNodes->findDisplayableNode(klickedOn->addr);

	sprintf(itemText, "Type '%s'", ((TypeNode*) dnd->node)->name);
	popup->setTitle(itemText);

	sprintf(itemText, "Show/Hide Type '%s'", ((TypeNode*) dnd->node)->name);
	popup->insertItem(itemText, parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));
	
	if ((dnd->flags & NODEFLAG_HIDDEN) && (dnd->flags & NODEFLAG_SHOWHIDDEN))
	  popup->insertItem("Omit This Hidden Type", 
			    this, SLOT(hideNodePopupSlot()));
	popup->insertSeparator();

	dn = dispNodes->findDisplayableNode(addrParent);
	QStringList names = namesOfHiddenChildNodes(addrParent);
	// Index 0 becomes the search item below, therefore start with index one
	for (int i = 0; i < names.count(); i++)
	  popup->insertItem(names[i], i + 1);
	popup->insertItem("Display All Hidden Types/Attributes on Level", 
			  this, SLOT(unhideLevelPopupSlot()));
	popup->insertItem("Omit All Hidden Types/Attributes on Level", 
			  this, SLOT(hideLevelPopupSlot()));

	popup->insertSeparator();

	sprintf(itemText, "Find Next Occurrence of '%s'", ((TypeNode*) dnd->node)->name);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((TypeNode*) dnd->node)->name,
									    klickedOn->addr);
	else if (index >= 1) {
	  // Index is one higher than the number of the child node
	  dn->node->unhideChildNode(this, addrParent, index - 1);
	  emit repaint();
	}

	break;
      }
      case NODETYPE_FEATVAL_NODE: {
	char itemText[500];
	NodeAddress addrParent;
	DisplayableNode* dnd;
	klickedOn->addr.parentAddress(addrParent);

	dn = dispNodes->findDisplayableNode(klickedOn->addr);
	dnd = dn;

	sprintf(itemText, "Attribute '%s'", ((FeatvalNode*) dnd->node)->feature);
	popup->setTitle(itemText);

	sprintf(itemText, "Show/Hide Attribute '%s'", ((FeatvalNode*) dnd->node)->feature);
	popup->insertItem(itemText, parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));
	
	if ((dn->flags & NODEFLAG_HIDDEN) && (dn->flags & NODEFLAG_SHOWHIDDEN))
	  popup->insertItem("Omit This Hidden Attribute", 
			    this, SLOT(hideNodePopupSlot()));
	popup->insertSeparator();

	dn = dispNodes->findDisplayableNode(addrParent);
	QStringList names = namesOfHiddenChildNodes(addrParent);

	// Index 0 becomes the search item below, therefore start with index one
	for (int i = 0; i < names.count(); i++)
	  popup->insertItem(names[i], i + 1);
	popup->insertItem("Display All Hidden Types/Attributes on Level", 
			  this, SLOT(unhideLevelPopupSlot()));
	popup->insertItem("Omit All Hidden Types/Attributes on Level", 
			  this, SLOT(hideLevelPopupSlot()));

	popup->insertSeparator();

	
	sprintf(itemText, "Find Next Occurrence of '%s'", ((FeatvalNode*) dnd->node)->feature);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((FeatvalNode*) dnd->node)->feature,
									    klickedOn->addr);
	else if (index >= 1) {
	  // Index is one higher than the number of the child node
	  dn->node->unhideChildNode(this, addrParent, index - 1);
	  emit repaint();
	}

	break;
      }
      case NODETYPE_LIST_NODE:
      case NODETYPE_TAIL_NODE:
	popup->setTitle("List");
	popup->insertItem("Show/Hide List", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));
	popup->popup(mapToGlobal(mo->pos()),2);
	break;
      case NODETYPE_SET_NODE:
      case NODETYPE_REST_NODE:
	popup->setTitle("Set");
	popup->insertItem("Show/Hide Set", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));
	popup->popup(mapToGlobal(mo->pos()),2);
	break;
      case NODETYPE_FUNCTION_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);

	popup->setTitle("Function");
	popup->insertItem("Show/Hide Function", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->insertSeparator();

	char itemText[500];
	sprintf(itemText, "Find Next Occurrence of '%s'", ((FunctionNode*) dn->node)->functor);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((FunctionNode*) dn->node)->functor, 
	    klickedOn->addr);

	break;
      }
      case NODETYPE_REF_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);

	popup->setTitle("Tag");
	if (!(dn->flags & NODEFLAG_EXPANDED))
	  popup->insertItem("Expand Tag", parentWidget()->parentWidget(), SLOT(expandLocalPopupSlot()));
	popup->insertItem("Show/Hide Tag", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->insertSeparator();

	char itemText[500];
	sprintf(itemText, "Find Next Occurrence of '%d'", ((RefNode*) dn->node)->ref);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0) {
	  char refStr[50];
	  sprintf(refStr, "%d", ((RefNode*) dn->node)->ref);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(refStr, klickedOn->addr);
	}

	break;
      }
      case NODETYPE_REENT_NODE:
	popup->setTitle("Reentrancy");
	popup->insertItem("Show/Hide Reentrancy", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));
	popup->popup(mapToGlobal(mo->pos()),2);
	break;
      case NODETYPE_TREE_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);

	popup->setTitle("Tree");
	popup->insertItem("Show/Hide Structure", this, SLOT(expandCollapseSlot()));
	popup->insertItem("Show/Hide Tree Node", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->insertSeparator();

	char itemText[500];
	sprintf(itemText, "Find Next Occurrence of '%s'", ((TreeNode*) dn->node)->label);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((TreeNode*) dn->node)->label, 
	    klickedOn->addr);

	break;
      }
      case NODETYPE_ATOM_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);

	popup->setTitle("Atom");
	popup->insertItem("Show/Hide Atom", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->insertSeparator();

	char itemText[500];
	sprintf(itemText, "Find Next Occurrence of '%s'", ((AtomNode*) dn->node)->atom);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((TreeNode*) dn->node)->label, 
	    klickedOn->addr);

	break;
      }
      case NODETYPE_REL_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);
	
	popup->setTitle("Relation");
	popup->insertItem("Show/Hide Relation", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->insertSeparator();

	char itemText[500];
	sprintf(itemText, "Find Next Occurrence of '%s'", ((RelationNode*) dn->node)->functor);
	popup->insertItem(QString::fromLocal8Bit(itemText), 0);

	int index = popup->exec(mapToGlobal(mo->pos()),2);

	if (index == 0)
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((RelationNode*) dn->node)->functor, 
	    klickedOn->addr);

	break;
      }
      case NODETYPE_STRUC_NODE: {
	dn = dispNodes->findDisplayableNode(klickedOn->addr);
	
	popup->setTitle("Feature Structure");
	popup->insertItem("Show/Hide Feature Structure", 
			  parentWidget()->parentWidget(), SLOT(hideUnhidePopupSlot()));

	popup->exec(mapToGlobal(mo->pos()),2);
	break;
      }
      }
    }
  }
  else if (mo->button() == LeftButton) {
    klickedOn = whereKlicked(mo->pos());
    if(klickedOn) {
      if (mo->state() & ShiftButton) {
	switch (klickedOn->type) {
	case NODETYPE_TYPE_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((TypeNode*) dn->node)->name, 
									    klickedOn->addr);
	  break;
	case NODETYPE_FEATVAL_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((FeatvalNode*) dn->node)->feature, 
									    klickedOn->addr);
	  break;
	case NODETYPE_FUNCTION_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((FunctionNode*) dn->node)->functor, 
									    klickedOn->addr);
	  break;
	case NODETYPE_REL_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((RelationNode*) dn->node)->functor, 
									    klickedOn->addr);
	  break;
	case NODETYPE_TREE_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((TreeNode*) dn->node)->label, 
									    klickedOn->addr);
	  break;
	case NODETYPE_REF_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);

	  if (mo->state() & ShiftButton) {
	    char refStr[50];
	    sprintf(refStr, "%d", ((RefNode*) dn->node)->ref);
	    ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(refStr, klickedOn->addr);
	  }

	  break;
	case NODETYPE_ATOM_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);
	  ((ViewCanvas*) parentWidget()->parentWidget())->contextMenuSearch(((AtomNode*) dn->node)->atom, 
									    klickedOn->addr);
	  break;
	}
      }
      else {
	switch (klickedOn->type) {
	case NODETYPE_TREE_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);

	  if (dn->flags & NODEFLAG_HIDDEN)
	    hideUnhide(dn);
	  else
	    expandCollapse(dn);
	  break;
	case NODETYPE_REF_NODE:
	  dn = dispNodes->findDisplayableNode(klickedOn->addr);

	  if (dn->flags & NODEFLAG_HIDDEN)
	    hideUnhide(dn);
	  else
	    expandCollapse(dn);
	  break;
	default:
	  // Uncomment to disable hiding of whole feature structures
	  //	  if (klickedOn->type != NODETYPE_STRUC_NODE) {
	    dn = dispNodes->findDisplayableNode(klickedOn->addr);
	    if (dn) {
	      if (dn->flags & NODEFLAG_HIDDEN)
		hideUnhide(dn);
	      else if (mo->state() & ControlButton)
		hideUnhide(dn);
	    }
	    //	  }
	}
      }
    }
  }
  if (mo->button() == MidButton) {
    dragX = mo->pos().x();
    dragY = mo->pos().y();
    dragging = true;
  }
}


void Canvas::mouseDoubleClickEvent(QMouseEvent *mo) {
  DisplayableNode* dn;

  if (mo->button() == LeftButton) {
    klickedOn = whereKlicked(mo->pos());
    if(klickedOn) {  
      dn = dispNodes->findDisplayableNode(klickedOn->addr);
      if ((dn->flags & NODEFLAG_HIDDEN) == 0)
	hideUnhide(dn);
    }
  }
}


void Canvas::mouseMoveEvent(QMouseEvent *mo) {
    int x = mo->pos().x();
    int y = mo->pos().y();
    
    /*    char buf[1000];
	  klickedOn = whereKlicked(mo->pos());
	  if (klickedOn) {
	  klickedOn->addr.sprint(buf);
	  ((ViewCanvas*) parentWidget()->parentWidget())->statusBar->message(buf);
	  }*/

    
    if (dragging) {
      int dx = dragX - x;
      int dy = dragY - y;

      dragX = x;  
      dragY = y;

      int nOx = offsetX + dx;
      int nOy = offsetY + dy;
      if(v_scroll->value() != nOy)
	v_scroll->setValue(nOy);
      if(h_scroll->value() != nOx)
	h_scroll->setValue(nOx);
    }
    else {
      klickedOn = whereKlicked(mo->pos());
      if (klickedOn != NULL) {
	DisplayableNode* nd = dispNodes->findDisplayableNode(klickedOn->addr);

	// If the node the user points on is a type node, featval node, or struc node, 
	// calculate a list containing the type and all
	// features of the structure node in order to display additional
	// information about the structure
	if ((nd->node->nodetype == NODETYPE_TYPE_NODE) ||
	    (nd->node->nodetype == NODETYPE_FEATVAL_NODE) ||
	    (nd->node->nodetype == NODETYPE_STRUC_NODE)) {
	  QStringList typeAndFeats;
	  NodeAddress addr;
	  StrucNode* snd;
	  
	  if (nd->node->nodetype == NODETYPE_STRUC_NODE)
	    addr = klickedOn->addr;
	  else
	    klickedOn->addr.parentAddress(addr);
	  
	  snd = (StrucNode*) (dispNodes->findDisplayableNode(addr)->node);
	  typeAndFeats = typeAndFeatsList(snd);
	  
	  QString statusMsg;
	  statusMsg.append(typeAndFeats[0]);
	  if (typeAndFeats.count() > 1) {
	    statusMsg.append(": ");
	    for (int i = 1; i < typeAndFeats.count(); i++) {
	      statusMsg.append(typeAndFeats[i]);
	      statusMsg.append(" ");
	    }
	  }
	  ((ViewCanvas*) parentWidget()->parentWidget())->setStatusBarMessage((const char*) statusMsg.local8Bit());
	}
	else
	  ((ViewCanvas*) parentWidget()->parentWidget())->setStatusBarMessage("");
	  
	if (options->highlightStructures) {
	  if (highlight == NULL) {
	    highlight = nd;
	    repaint(true);
	  }
	  else if (highlight != nd) {
	    highlight = nd;
	    repaint(true);
	  }
	}
      }
      else {
	((ViewCanvas*) parentWidget()->parentWidget())->setStatusBarMessage("");
	if (options->highlightStructures) {
	  if (highlight != NULL) {
	    highlight = NULL;
	    repaint(true);
	  }
	}
      }
    }
}


void Canvas::mouseReleaseEvent(QMouseEvent *mo) {
  if (mo->button() == MidButton) {
    dragging = false;
  }
}


Canvas::KlickBox* Canvas::whereKlicked(const QPoint &p) {
  KlickBox *klick;
  for(klick = toKlick.first(); klick != NULL; klick = toKlick.next()) {
    if (klick->box.contains(p)) {
      return klick;
    }
  }

  return NULL;
}



void Canvas::hideUnhide(DisplayableNode* nd, bool redraw) {
  if (nd->flags & NODEFLAG_HIDDEN)
    nd->flags &= ~NODEFLAG_HIDDEN;
  else
    nd->flags |= NODEFLAG_HIDDEN;
 
  // repaint ...
  if ( redraw ) {
    erase();
    emit redrawSignal();
  }
}


void Canvas::hideLevelPopupSlot(void) {
  NodeAddress addrParent;
  klickedOn->addr.parentAddress(addrParent);
  
  DisplayableNode* dn = dispNodes->findDisplayableNode(addrParent);
  if (dn) {
    dn->node->hideUnhideChildNodes(this, addrParent, true);
    emit redrawSignal();
  }
}


void Canvas::unhideLevelPopupSlot(void) {
  NodeAddress addrParent;
  klickedOn->addr.parentAddress(addrParent);
  
  DisplayableNode* dn = dispNodes->findDisplayableNode(addrParent);
  if (dn) {
    dn->node->hideUnhideChildNodes(this, addrParent, false);
    emit redrawSignal();
  }
}


void Canvas::hideNodePopupSlot(void) {
  DisplayableNode* dn = dispNodes->findDisplayableNode(klickedOn->addr);
  if (dn) {
    dn->flags &= ~NODEFLAG_SHOWHIDDEN;
    emit redrawSignal();
  }
}
  


void Canvas::expandCollapse(DisplayableNode* nd, bool redraw) {
  if (nd->flags & NODEFLAG_EXPANDED)
    nd->flags &= ~NODEFLAG_EXPANDED;
  else
    nd->flags |= NODEFLAG_EXPANDED;

  if (redraw) {
    erase();
    emit redrawSignal();
  }
}

void Canvas::expandCollapseSlot() {
  expandCollapse(dispNodes->findDisplayableNode(klickedOn->addr), true);
}

void Canvas::setRootNode(GenericNode* node) {
  if (node)
    rootNode = node;
  else
    rootNode = document->data->at(0);

  highlight = NULL;
  foundMark = NULL;

  emit redrawSignal();
}

void Canvas::resetScrollbar()  { 
  h_scroll->setValue(offsetX);
  h_scroll->repaint();
  v_scroll->setValue(offsetY);
  v_scroll->repaint();
}

Canvas::KlickBox* Canvas::getKlickedOn() {
  return klickedOn;
}

IdTree* Canvas::getDispNodes(void) {
  return dispNodes;
}

void Canvas::reset() {
  delete dispNodes;
  dispNodes = new IdTree(document->maxID() + 1);
  emit redrawSignal();
  highlight = NULL;
  foundMark = NULL;
}
  
void Canvas::resetToClick(void) {
  toKlick.clear();
}

void Canvas::updateOptions(const Options& op) {
  options->backgroundColor = op.backgroundColor;
  options->normalColor = op.normalColor;
  options->typeColor = op.typeColor;
  options->textColor = op.textColor;
  options->errorColor = op.errorColor;
  options->treeColor = op.treeColor;
  options->treeLineColor = op.treeLineColor;
  options->leafColor = op.leafColor;
  options->hiddenColor = op.hiddenColor;
  options->markerColor = op.markerColor;
  options->foundColor = op.foundColor;
  options->differentNodeColor = op.differentNodeColor;
  options->arclabelColor = op.arclabelColor;
  options->arclabelBgColor = op.arclabelBgColor;

  options->normalFont = op.normalFont;
  options->typeFont = op.typeFont;
  options->textFont = op.textFont;
  options->treeFont = op.treeFont;
  options->leafFont = op.leafFont;
  options->errorFont = op.errorFont;
  options->arclabelFont = op.arclabelFont;
}
  
void Canvas::ensureVisible(const NodeAddress* addr) {
  char id[ID_LENGTH];
  int idpos = 0;
  char idb[ID_LENGTH];
  int i;
  NodeAddress a;
  DisplayableNode* dnd;
  bool expand;
  GenericNode* node;
  for (int pos = 0; pos < addr->len; pos++) {
    id[pos] = (addr->addr[pos] & 0x0F) + '0';
    if (addr->addr[pos] & 0x40) {
      pos++;
      id[pos] = 0x00;
      break;
    }
  }
  node = document->findNodeById(id);
  if (node->nodetype == NODETYPE_TREE_NODE)
    expand = false;
  else
    expand = true;
    
  for (int pos = addr->len - 1; pos >= 0; pos--) {
    idb[idpos] = (addr->addr[pos] & 0x0F) + '0';
    if (addr->addr[pos] & 0x80) {
      for (i = 0; i <= idpos; i++)
	id[i] = idb[idpos - i];
      id[i] = 0x00;
      NodeAddress ad(a, id);
      node = document->findNodeById(id);
      dnd = dispNodes->addDisplayableNode(ad, node);
      dnd->flags &= ~(NODEFLAG_HIDDEN);
      if (expand && (pos > 0))
	dnd->flags |= NODEFLAG_EXPANDED;
      a = ad;
      idpos = 0;
    }
    else
      idpos++;
  }
}

QPoint Canvas::nodeOrigin(const NodeAddress* addr) {
  KlickBox *klick;
  for(klick = toKlick.first(); klick != NULL; klick = toKlick.next()) {
    if (klick->addr == *addr)
      return QPoint((klick->box.left() + offsetX > 20 ? klick->box.left() + offsetX - 20 : 0), (klick->box.top() + offsetY > 20 ? klick->box.top() + offsetY - 20 : 0));
  }
  return QPoint(0, 0);
}


      
void Canvas::markFound(const NodeAddress* addr) {
  DisplayableNode* dnd;
  if (foundMark != NULL)
    foundMark->flags &= ~(NODEFLAG_FOUND);
  if (addr == NULL) {
    foundMark = NULL;
    repaint(false);
  }
  else {
    dnd = dispNodes->findDisplayableNode(*addr);
    dnd->flags |= NODEFLAG_FOUND;
    foundMark = dnd;

    repaint(false);
    QPoint org = nodeOrigin(addr);
    
    if (v_scroll->value() != org.y())
      v_scroll->setValue(org.y());
    if (h_scroll->value() != org.x())
      h_scroll->setValue(org.x());
  }
}


Document* Canvas::getDocument(void) const {
  return document;
}


/** Returns a list that contains the type and all attributes of a
    feature structure (a <code>StrucNode</code>).
    @param node the <code>StrucNode</code>
    @return a list containing the type as its first element followed by all
      attributes. */
QStringList Canvas::typeAndFeatsList(StrucNode* node) {
  QStringList list;
  List* featvals;
  
  list.append(QString::fromLocal8Bit(node->type->name));

  featvals = node->featvals;
  while (featvals) {
    list.append(QString::fromLocal8Bit(((FeatvalNode*) featvals->item)->feature));
    featvals = featvals->tail;
  }

  return list;
}

  
QStringList Canvas::namesOfHiddenChildNodes(NodeAddress& addr) {
  QStringList names;
  DisplayableNode* dnd = dispNodes->findDisplayableNode(addr);
  char name[500];
  
  if (dnd) {
    if (dnd->node->nodetype == NODETYPE_STRUC_NODE) {
      StrucNode* nd = (StrucNode*) dnd->node;
      List* featvals;
      
      dnd = dispNodes->addDisplayableNode(NodeAddress(addr, nd->type), nd->type);
      if ((dnd->flags & NODEFLAG_HIDDEN) && !(dnd->flags & NODEFLAG_SHOWHIDDEN)) {
	sprintf(name, "Show Type: %s", nd->type->name);
	names.append(QString::fromLocal8Bit(name));
      }
      
      featvals = nd->featvals;
      while (featvals) {
	dnd = dispNodes->addDisplayableNode(NodeAddress(addr, featvals->item), featvals->item);
	if ((dnd->flags & NODEFLAG_HIDDEN) && !(dnd->flags & NODEFLAG_SHOWHIDDEN)) {
	  sprintf(name, "Show Attribute: %s", ((FeatvalNode*) featvals->item)->feature);
	  names.append(QString::fromLocal8Bit(name));
	}
	featvals = featvals->tail;
      }
    }
  }

  return names;
}
