/*  $Id: compare.cpp,v 1.29 2003/05/22 16:43:09 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Grisu text-based structure compare algorithm.

    WARNING!!! Since Grisu is mainly used as the frontend for the TRALE system, which has
    built-in structure comparison functionality, this algorithm has NOT BEEN extended
    to support newer node types like NODETYPE_DISJUNCTION_NODE. If you want to use this
    code, you MUST add the appropriate handling of these node types.

    @author Holger Wunsch */

#include "data.h"
#include "compare.h"
#include "dispnode.h"
#include <string.h>

StructureCompare::StructureCompare(Document* document1, const NodeAddress& addr1, GenericNode* node1, IdTree* tree1, Document* document2, const NodeAddress& addr2, GenericNode* node2, IdTree* tree2):
  ad1(addr1), ad2(addr2){
  this->document1 = document1;
  this->nd1 = node1;
  this->tree1 = tree1;

  this->document2 = document2;
  this->nd2 = node2;
  this->tree2 = tree2;

  cyclic = false;
  diffs = 0;
}

void StructureCompare::markAsDifferent(const NodeAddress& addr, GenericNode* node, IdTree* tree) {
  DisplayableNode* dnd = tree->addDisplayableNode(addr, node);
  dnd->flags |= NODEFLAG_DIFFERENT;
}

StructureCompare::~StructureCompare() {
}


void StructureCompare::compare() {
  compareGenericNodes(ad1, nd1, ad2, nd2);
}

void StructureCompare::compareGenericNodes(const NodeAddress& parentAddr1, GenericNode* node1, const NodeAddress& parentAddr2, GenericNode* node2) {
  NodeAddress addr1(parentAddr1, node1);
  NodeAddress addr2(parentAddr2, node2);

  if ((node1->nodetype == NODETYPE_TYPE_NODE) && (node2->nodetype == NODETYPE_TYPE_NODE))
    compareTypeNodes(parentAddr1, (TypeNode*) node1, parentAddr2, (TypeNode*) node2);
  else if ((node1->nodetype == NODETYPE_STRUC_NODE) && (node2->nodetype == NODETYPE_STRUC_NODE))
    compareStrucNodes(parentAddr1, (StrucNode*) node1, parentAddr2, (StrucNode*) node2);
  else if ((node1->nodetype == NODETYPE_FEATVAL_NODE) && (node2->nodetype == NODETYPE_FEATVAL_NODE))
    compareFeatvalNodes(parentAddr1, (FeatvalNode*) node1, parentAddr2, (FeatvalNode*) node2);
  else if ((node1->nodetype == NODETYPE_LIST_NODE) && (node2->nodetype == NODETYPE_LIST_NODE))
    compareListNodes(parentAddr1, (ListNode*) node1, parentAddr2, (ListNode*) node2);
  else if ((node1->nodetype == NODETYPE_SET_NODE) && (node2->nodetype == NODETYPE_SET_NODE))
    compareSetNodes(parentAddr1, (SetNode*) node1, parentAddr2, (SetNode*) node2);
  else if ((node1->nodetype == NODETYPE_FUNCTION_NODE) && (node2->nodetype == NODETYPE_FUNCTION_NODE))
    compareFunctionNodes(parentAddr1, (FunctionNode*) node1, parentAddr2, (FunctionNode*) node2);
  else if ((node1->nodetype == NODETYPE_REF_NODE) && (node2->nodetype == NODETYPE_REF_NODE))
    compareRefNodes(parentAddr1, (RefNode*) node1, parentAddr2, (RefNode*) node2);
  else if ((node1->nodetype == NODETYPE_REENT_NODE) && (node2->nodetype == NODETYPE_REENT_NODE))
    compareReentNodes(parentAddr1, (ReentNode*) node1, parentAddr2, (ReentNode*) node2);
  else if ((node1->nodetype == NODETYPE_TREE_NODE) && (node2->nodetype == NODETYPE_TREE_NODE))
    compareTreeNodes(parentAddr1, (TreeNode*) node1, parentAddr2, (TreeNode*) node2);
  else if ((node1->nodetype == NODETYPE_ATOM_NODE) && (node2->nodetype == NODETYPE_ATOM_NODE))
    compareAtomNodes(parentAddr1, (AtomNode*) node1, parentAddr2, (AtomNode*) node2);
  else if ((node1->nodetype == NODETYPE_REL_NODE) && (node2->nodetype == NODETYPE_REL_NODE))
    compareRelationNodes(parentAddr1, (RelationNode*) node1, parentAddr2, (RelationNode*) node2);
  else if ((node1->nodetype == NODETYPE_TAIL_NODE) && (node2->nodetype == NODETYPE_TAIL_NODE))
    compareTailNodes(parentAddr1, (TailNode*) node1, parentAddr2, (TailNode*) node2);
  else if ((node1->nodetype == NODETYPE_REST_NODE) && (node2->nodetype == NODETYPE_REST_NODE))
    compareRestNodes(parentAddr1, (RestNode*) node1, parentAddr2, (RestNode*) node2);
  else if ((node1->nodetype == NODETYPE_REF_NODE) && (node2->nodetype != NODETYPE_REF_NODE)) {
    ReentNode* reentNode = document1->findReentrancy(((RefNode*) node1)->ref);
    compareGenericNodes(addr1, reentNode->value, parentAddr2, node2);
    markAsDifferent(addr1, node1, tree1);
    diffs++;
  }
  else if ((node1->nodetype != NODETYPE_REF_NODE) && (node2->nodetype == NODETYPE_REENT_NODE)) {
    ReentNode* reentNode = document2->findReentrancy(((RefNode*) node2)->ref);
    compareGenericNodes(parentAddr1, node1, addr2, reentNode->value);
    markAsDifferent(addr2, node2, tree2);
    diffs++;
  }
  else {
    markAsDifferent(addr1, node1, tree1);
    markAsDifferent(addr2, node2, tree2);
    diffs++;
  } 
}

void StructureCompare::compareTypeNodes(const NodeAddress& parentAddr1, TypeNode* node1, const NodeAddress& parentAddr2, TypeNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);
  if (strcmp(node1->name, node2->name) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }
}

void StructureCompare::compareFeatvalNodes(const NodeAddress& parentAddr1, FeatvalNode* node1, const NodeAddress& parentAddr2, FeatvalNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);  
  if (strcmp(node1->feature, node2->feature) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }
  else
    compareGenericNodes(addr1, node1->value, addr2, node2->value);
}   
    
void StructureCompare::compareStrucNodes(const NodeAddress& parentAddr1, StrucNode* node1, const NodeAddress& parentAddr2, StrucNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);
  compareTypeNodes(addr1, node1->type, addr2, node2->type);
  
  List* featvals1 = node1->featvals;
  List* featvals2 = node2->featvals;

  while ((featvals1 != NULL) || (featvals2 != NULL)) {
    if ((featvals1 != NULL) && (featvals2 != NULL)){
      compareGenericNodes(addr1, featvals1->item, addr2, featvals2->item);
      featvals1 = featvals1->tail;
      featvals2 = featvals2->tail;
    }
    else if (featvals1 == NULL) {
      markAsDifferent(NodeAddress(addr2, featvals2->item), featvals2->item, tree2);
      diffs++;
      featvals2 = featvals2->tail;
    }
    else if (featvals2 == NULL) {
      markAsDifferent(NodeAddress(addr1, featvals1->item), featvals1->item, tree1);
      diffs++;
      featvals1 = featvals1->tail;
    }
  }
}

void StructureCompare::compareTailNodes(const NodeAddress& parentAddr1, TailNode* node1, const NodeAddress& parentAddr2, TailNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);
  compareGenericNodes(addr1, node1->tail, addr2, node2->tail);
}

void StructureCompare::compareListNodes(const NodeAddress& parentAddr1, ListNode* node1, const NodeAddress& parentAddr2, ListNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);

  List* structs1 = node1->structs;
  List* structs2 = node2->structs;

  while ((structs1 != NULL) || (structs2 != NULL)) {
    if ((structs1 != NULL) && (structs2 != NULL)) {
      compareGenericNodes(addr1, structs1->item, addr2, structs2->item);
      structs1 = structs1->tail;
      structs2 = structs2->tail;
    }
    else if (structs1 == NULL) {
      markAsDifferent(NodeAddress(addr2, structs2->item), structs2->item, tree2);
      diffs++;
      structs2 = structs2->tail;
    }
    else if (structs2 == NULL) {
      markAsDifferent(NodeAddress(addr1, structs1->item), structs1->item, tree1);
      diffs++;
      structs1 = structs1->tail;
    }
  }

  if ((node1->tail != NULL) && (node2->tail != NULL))
    compareTailNodes(addr1, node1->tail, addr2, node2->tail);
}

void StructureCompare::compareSetNodes(const NodeAddress& parentAddr1, SetNode* node1, const NodeAddress& parentAddr2, SetNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);

  List* structs1 = node1->structs;
  List* structs2 = node2->structs;

  while ((structs1 != NULL) || (structs2 != NULL)) {
    if ((structs1 != NULL) && (structs2 != NULL)) {
      compareGenericNodes(addr1, structs1->item, addr2, structs2->item);
      structs1 = structs1->tail;
      structs2 = structs2->tail;
    }
    else if (structs1 == NULL) {
      markAsDifferent(NodeAddress(addr2, structs2->item), structs2->item, tree2);
      diffs++;
      structs2 = structs2->tail;
    }
    else if (structs2 == NULL) {
      markAsDifferent(NodeAddress(addr1, structs1->item), structs1->item, tree1);
      diffs++;
      structs1 = structs1->tail;
    }
  }

  if ((node1->rest != NULL) && (node2->rest != NULL))
    compareRestNodes(addr1, node1->rest, addr2, node2->rest);
}

void StructureCompare::compareRestNodes(const NodeAddress& parentAddr1, RestNode* node1, const NodeAddress& parentAddr2, RestNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);
  compareGenericNodes(addr1, node1->rest, addr2, node2->rest);
}

void StructureCompare::compareFunctionNodes(const NodeAddress& parentAddr1, FunctionNode* node1, const NodeAddress& parentAddr2, FunctionNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);

  compareTypeNodes(addr1, node1->type, addr2, node2->type);

  if (strcmp(node1->functor, node2->functor) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }

  List* structs1 = node1->structs;
  List* structs2 = node2->structs;
  while ((structs1 != NULL) || (structs2 != NULL)) {
    if ((structs1 != NULL) && (structs2 != NULL)) {
      compareGenericNodes(addr1, structs1->item, addr2, structs2->item);
      structs1 = structs1->tail;
      structs2 = structs2->tail;
    }
    else if (structs1 == NULL) {
      markAsDifferent(NodeAddress(addr2, structs2->item), structs2->item, tree2);
      diffs++;
      structs2 = structs2->tail;
    }
    else if (structs2 == NULL) {
      markAsDifferent(NodeAddress(addr1, structs1->item), structs1->item, tree1);
      diffs++;
      structs1 = structs1->tail;
    }
  }
}

void StructureCompare::compareRelationNodes(const NodeAddress& parentAddr1, RelationNode* node1, const NodeAddress& parentAddr2, RelationNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);

  if (strcmp(node1->functor, node2->functor) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }

  List* structs1 = node1->structs;
  List* structs2 = node2->structs;
  while ((structs1 != NULL) || (structs2 != NULL)) {
    if ((structs1 != NULL) && (structs2 != NULL)) {
      compareGenericNodes(addr1, structs1->item, addr2, structs2->item);
      structs1 = structs1->tail;
      structs2 = structs2->tail;
    }
    else if (structs1 == NULL) {
      markAsDifferent(NodeAddress(addr2, structs2->item), structs2->item, tree2);
      diffs++;
      structs2 = structs2->tail;
    }
    else if (structs2 == NULL) {
      markAsDifferent(NodeAddress(addr1, structs1->item), structs1->item, tree1);
      diffs++;
      structs1 = structs1->tail;
    }
  }
}
 
void StructureCompare::compareRefNodes(const NodeAddress& parentAddr1, RefNode* node1, const NodeAddress& parentAddr2, RefNode* node2) {
  if (parentAddr1.contains(node1->id) || parentAddr2.contains(node2->id)) {
    cyclic = true;
    return;
  }
  
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*)  node2);

  compareReentNodes(addr1, document1->findReentrancy(node1->ref), addr2, document2->findReentrancy(node2->ref));
}
 
void StructureCompare::compareReentNodes(const NodeAddress& parentAddr1, ReentNode* node1, const NodeAddress& parentAddr2, ReentNode* node2) {
  compareGenericNodes(parentAddr1, node1->value, parentAddr2, node2->value);
}

void StructureCompare::compareTreeNodes(const NodeAddress& parentAddr1, TreeNode* node1, const NodeAddress& parentAddr2, TreeNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);

  if (strcmp(node1->label, node2->label) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }
  
  GenericNode* linked1 = document1->findNodeById(node1->linkid);
  GenericNode* linked2 = document2->findNodeById(node2->linkid);

  if ((linked1 != NULL) && (linked2 != NULL))
    compareGenericNodes(addr1, document1->findNodeById(node1->linkid), addr2, document2->findNodeById(node2->linkid));
  
  List* subtrees1 = node1->subtrees;
  List* subtrees2 = node2->subtrees;

  while ((subtrees1 != NULL) || (subtrees2 != NULL)) {
    if ((subtrees1 != NULL) && (subtrees2 != NULL)) {
      compareGenericNodes(addr1, subtrees1->item, addr2, subtrees2->item);
      subtrees1 = subtrees1->tail;
      subtrees2 = subtrees2->tail;
    }
    else if (subtrees1 == NULL) {
     markAsDifferent(NodeAddress(addr2, subtrees2->item), subtrees2->item, tree2);
     diffs++;
     subtrees2 = subtrees2->tail;
    }
    else if (subtrees2 == NULL) {
      markAsDifferent(NodeAddress(addr1, subtrees1->item), subtrees1->item, tree1);
      diffs++;
      subtrees1 = subtrees1->tail;
    }
  }
}
 
void StructureCompare::compareAtomNodes(const NodeAddress& parentAddr1, AtomNode* node1, const NodeAddress& parentAddr2, AtomNode* node2) {
  NodeAddress addr1(parentAddr1, (GenericNode*) node1);
  NodeAddress addr2(parentAddr2, (GenericNode*) node2);
  
  if (strcmp(node1->atom, node2->atom) != 0) {
    markAsDifferent(addr1, (GenericNode*) node1, tree1);
    markAsDifferent(addr2, (GenericNode*) node2, tree2);
    diffs++;
  }
}

bool StructureCompare::cyclicPaths() {
  return cyclic;
}

int StructureCompare::differences() {
  return diffs;
}

