/*  $Id: find.cpp,v 1.29 2003/05/22 16:43:11 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Implementation of the <code>Finder</code> class.

    @author Holger Wunsch */

#include <qmessagebox.h>
#include "find.h"

Finder::Finder(Document* doc, QArray<int>* typefilter, const char* argument, bool ignoreCase) {
  d = doc;
  foundList = new QList<NodeAddress>;
  foundList->setAutoDelete(true);
  this->typefilter = typefilter;
  this->argument = new char[strlen(argument) + 1];
  this->ignoreCase = ignoreCase;
  strcpy(this->argument, argument);
}

Finder::~Finder(void) {
  delete [] argument;
  delete foundList;
}

void Finder::findGenericNode(GenericNode* node, const NodeAddress& addrParent) {
  switch (node->nodetype) {
  case NODETYPE_TYPE_NODE:
    findTypeNode((TypeNode*) node, addrParent);
    break;
  case NODETYPE_STRUC_NODE:
    findStrucNode((StrucNode*) node, addrParent);
    break;
  case NODETYPE_FEATVAL_NODE:
    findFeatvalNode((FeatvalNode*) node, addrParent);
    break;
  case NODETYPE_LIST_NODE:
    findListNode((ListNode*) node, addrParent);
    break;
  case NODETYPE_SET_NODE:
    findSetNode((SetNode*) node, addrParent);
    break;
  case NODETYPE_FUNCTION_NODE:
    findFunctionNode((FunctionNode*) node, addrParent);
    break;
  case NODETYPE_REF_NODE:
    findRefNode((RefNode*) node, addrParent);
    break;
  case NODETYPE_REENT_NODE:
    findReentNode((ReentNode*) node, addrParent);
    break;
  case NODETYPE_TREE_NODE:
    findTreeNode((TreeNode*) node, addrParent);
    break;
  case NODETYPE_ATOM_NODE:
    findAtomNode((AtomNode*) node, addrParent);
    break;
  case NODETYPE_REL_NODE:
    findRelNode((RelationNode*) node, addrParent);
    break;
  case NODETYPE_TAIL_NODE:
    findTailNode((TailNode*) node, addrParent);
    break;
  case NODETYPE_REST_NODE:
    findRestNode((RestNode*) node, addrParent);
    break;
  case NODETYPE_DISJUNCTION_NODE:
    findDisjunctionNode((DisjunctionNode*) node, addrParent);
    break;
  case NODETYPE_CONJUNCTION_NODE:
    findConjunctionNode((ConjunctionNode*) node, addrParent);
    break;
  default:
    printf("Invalid nodetype in find: %d\n", node->nodetype);
    break;
  }
}

void Finder::findTypeNode(TypeNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_TYPE_NODE) > -1) {
    if (strCmp(node->name, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
}

void Finder::findStrucNode(StrucNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  List* fv = node->featvals;
  
  findTypeNode(node->type, addr);
  while (fv) {
    findGenericNode(fv->item, addr);
    fv = fv->tail;
  }
}

void Finder::findFeatvalNode(FeatvalNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_FEATVAL_NODE) > -1) {
    if (strCmp(node->feature, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
  findGenericNode(node->value, addr);
}

void Finder::findListNode(ListNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  List* structs = node->structs;
  while (structs) {
    findGenericNode(structs->item, addr);
    structs = structs->tail;
  }
  if (node->tail != NULL)
    findTailNode(node->tail, addr);
}

void Finder::findSetNode(SetNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  List* structs = node->structs;
  while (structs) {
    findGenericNode(structs->item, addr);
    structs = structs->tail;
  }
  if (node->rest != NULL)
    findRestNode(node->rest, addr);
}

void Finder::findFunctionNode(FunctionNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  findTypeNode(node->type, addr);
  if (typefilter->find(NODETYPE_FUNCTION_NODE) > -1) {
    if (strCmp(node->functor, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
  List* structs = node->structs;
  while(structs) {
    findGenericNode(structs->item, addr);
    structs = structs->tail;
  }
}
    
void Finder::findRefNode(RefNode* node, const NodeAddress& addrParent) {
  if (addrParent.contains(node->id)) {
    //QMessageBox::warning(NULL, "Warning", "Cyclic structures detected. Search stopped at cycle.");
    return;
  }
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_REF_NODE) > -1) {
    char refStr[20];
    sprintf(refStr, "%d", node->ref);
    if (strCmp(refStr, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
  ReentNode* reent = d->findReentrancy(node->ref);
  findReentNode(reent, addr);
}

void Finder::findReentNode(ReentNode* node, const NodeAddress& addrParent) {
  findGenericNode(node->value, addrParent);
}

void Finder::findTreeNode(TreeNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_TREE_NODE) > -1) {
    if (strCmp(node->label, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
  GenericNode* nd = d->findNodeById(node->linkid);
  if (nd)  // Leaves have no associated nodes!
    findGenericNode(nd, addr);
  List* subtrees = node->subtrees;
  while (subtrees) {
    findTreeNode((TreeNode*) subtrees->item, addr);
    subtrees = subtrees->tail;
  }
}

void Finder::findAtomNode(AtomNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_ATOM_NODE) > -1) {
    if (strCmp(node->atom, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
}

void Finder::findRelNode(RelationNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  if (typefilter->find(NODETYPE_REL_NODE) > -1) {
    if (strCmp(node->functor, argument) == 0)
      foundList->append(new NodeAddress(addr));
  }
  List* structs = node->structs;
  while (structs) {
    findGenericNode(structs->item, addr);
    structs = structs->tail;
  }
}

void Finder::findTailNode(TailNode* node, const NodeAddress& addrParent) {
  findGenericNode(node->tail, addrParent);
}

void Finder::findRestNode(RestNode* node, const NodeAddress& addrParent) {
  findGenericNode(node->rest, addrParent);
}


void Finder::findDisjunctionNode(DisjunctionNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  List* operands = node->operands;
  while (operands) {
    findGenericNode(operands->item, addr);
    operands = operands->tail;
  }
}


void Finder::findConjunctionNode(ConjunctionNode* node, const NodeAddress& addrParent) {
  NodeAddress addr(addrParent, (GenericNode*) node);
  List* operands = node->operands;
  while (operands) {
    findGenericNode(operands->item, addr);
    operands = operands->tail;
  }
}


const NodeAddress* Finder::first() {
  return foundList->first();
}

const NodeAddress* Finder::next() {
  return foundList->next();
}


int Finder::strCmp(const char* s1, const char* s2) {
  if (ignoreCase)
    return strcasecmp(s1, s2);
  else
    return strcmp(s1, s2);
}
