/*  $Id: find.h,v 1.29 2003/05/22 16:43:11 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    Declarations for the <code>Finder</code> class.

    @author Holger Wunsch */

#ifndef __FIND_H
#define __FIND_H

#include <qstring.h>
#include <qlist.h>
#include <qarray.h>
#include "data.h"
#include "node.h"
#include "NodeAddress.h"


/** This class provides the functionality to search for a given
    search term in a data structure. The node addresses of all 
    nodes that match the search term are kept in a list. Using the
    built-in type filter, it is possible to selectively include and
    exclude nodes in or from the search based on their type.

    @author Holger Wunsch */
class Finder {
  /** The list of addresses of nodes that match the search term. */
  QList<NodeAddress>* foundList;


  /** The search term. */
  char* argument;


  /** An array that contains all types of nodes to be included in
      the next search. Valid values for array elements are the
      <code>NODETYPE_xxxx</code> constants. */
  QArray<int>* typefilter;


  /** The document that contains the data structure to be searched. */
  Document* d;


  /** <code>true</code> if all string comparisons should ignore case,
      <code>false</code> otherwise. */
  bool ignoreCase;


  /** Searches in a TypeNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findTypeNode(TypeNode* node, const NodeAddress& addrParent);


  /** Searches in a StrucNode. 
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findStrucNode(StrucNode* node, const NodeAddress& addrParent);


  /** Searches in a FeatvalNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findFeatvalNode(FeatvalNode* node, const NodeAddress& addrParent);


  /** Searches in a ListNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findListNode(ListNode* node, const NodeAddress& addrParent);


  /** Searches in a SetNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findSetNode(SetNode* node, const NodeAddress& addrParent);


  /** Searches in a FunctionNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findFunctionNode(FunctionNode* node, const NodeAddress& addrParent);


  /** Searches in a RefNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findRefNode(RefNode* node, const NodeAddress& addrParent);


  /** Searches in a ReentNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findReentNode(ReentNode* node, const NodeAddress& addrParent);


  /** Searches in a TreeNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findTreeNode(TreeNode* node, const NodeAddress& addrParent);


  /** Searches in a AtomNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findAtomNode(AtomNode* node, const NodeAddress& addrParent);


  /** Searches in a RelNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findRelNode(RelationNode* node, const NodeAddress& addrParent);


  /** Searches in a TailNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findTailNode(TailNode* node, const NodeAddress& addrParent);


  /** Searches in a RestNode.
      @param node the node
      @param addrParent the address of this node's parent node. */
  void findRestNode(RestNode* node, const NodeAddress& addrParent);

  
  /** Searches within a DisjunctionNode.
      @param node the DisjunctionNode
      @param addrParent the node address of the displayable parent node. */
  void findDisjunctionNode(DisjunctionNode* node, const NodeAddress& addrParent);


  /** Searches within a ConjunctionNode.
      @param node the ConjunctionNode
      @param addrParent the node address of the displayable parent node. */
  void findConjunctionNode(ConjunctionNode* node, const NodeAddress& addrParent);


  /** String compare that ignores case in the comparison depending on the
      value of <code>ignoreCase</code>.
      @param s1 the first string
      @param s2 the second string
      @return the result of the comparison (either 0, < 0, or > 0; see the documentation
      of the <code>strcmp</code> and <code>strcasecmp</code> C-library routines). */
  int strCmp(const char* s1, const char* s2);
  
 public:
  /** Creates a new finder with a new search term.
      @param doc the document which contains the data to be searched.
      @param typefilter an array that contains all types of nodes that should be included
        in this search. The values of each element in the array must be set to one of
	the <code>NODETYPE_xxxx</code> constants.
      @param argument the search term, an alphanumeric string
      @param ignoreCase <code>true</code> makes the finder ignore case */
  Finder(Document* doc, QArray<int>* typefilter, const char* argument, bool ignoreCase);


  /** Deletes this instance.*/
  ~Finder(void);


  /** Starts the search at the given node. After calling this function, the addresses of
      the found nodes can be retrieved with <code>first()</code> and <code>next()</code>.
      @param node the node where the search is started
      @param addrParent this node's parent address. */
  void findGenericNode(GenericNode* node, const NodeAddress& addrParent);


  /** Retrieves the first node in the found list.
      @return the first node in the found list, or <code>NULL</code> if no node was found
        at all. */
  const NodeAddress* first();


  /** Retrieves the next in the found list.
      @return the next node in the found list, or <code>NULL</code> if no more nodes
      are in the found list. */
  const NodeAddress* next();
};

#endif


  

  
