/*  $Id: gdataexchange.h,v 1.12 2003/05/22 16:43:11 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/** @file

    Declaration of the GDataExchange class.

    @author Holger Wunsch */


#ifndef __GDATAEXCHANGE_H
#define __GDATAEXCHANGE_H


#include <qobject.h>
#include <ksock.h>
#include "gbuffer.h"
#include "gdatareader.h"


/** The initial size of the outbound data buffer. */
#define GDE_OUTBOUND_BUFFER_SIZE   1000


/** The initial size of the inbound data buffer. */
#define GDE_INBOUND_BUFFER_SIZE    10000


/** The amount the inbound data buffer grows if necessary. */
#define GDE_INBOUND_BUFFER_GROW_DELTA   10000


/** The low end of the range of port numbers Grisu will accept connections on. */
#define GDE_PORTRANGE_LO   5001


/** The high end of the range of port numbers Grisu will accept connections on. */
#define GDE_PORTRANGE_HI   5050


/** The maximum length of a hostname. */
#define GDE_HOSTNAME_MAXLEN  500


/* Connection states. */
#define GDE_STATE_NO_CONNECTION  0  /**< Connection state: Connection not possible */
#define GDE_STATE_READY          1  /**< Connection state: Ready for connection */
#define GDE_STATE_CONNECTED      2  /**< Connection state: Connected. */


/** This class implements Grisu's bidirectional data exchange interface
    based on TCP/IP sockets.

    GDataExchange provides both asynchronous and synchronous I/O on the
    socket connection. 

    In asnychronous mode, when incoming data becomes available, this data
    is read without any interaction from outside this class. After reading
    has completed, a signal is raised which contains the data as an argument.
    
    Outbound data is buffered within GDataExchange and then sent out
    asynchronously.

    @author Holger Wunsch */
class GDataExchange: public QObject {
  Q_OBJECT

  /** The server socket, which controls all connections to clients. */
  KServerSocket* serverSocket;


  /** The socket that is used in the current connection. */
  KSocket* socket;

  
  /** The outbound data buffer. */
  GBuffer* outData;


  /** The inbound data buffer. */
  GBuffer* inData;


  /** Pointer to a class that handles incoming data. */
  GDataReader* dataReader;


  /** The name of the host Grisu is running on. */
  char hostname[GDE_HOSTNAME_MAXLEN];


  /** The number of the port that is used in the current connection
      (<code>-1</code> if no connection has been established so far). */
  int port;


  /** Closes the server socket. */
  void closeServerSocket(void);


 public:
  /** Creates a new data exchange object. Sets up the server socket.
      @param reader the data reader that handles incoming data. */
  GDataExchange(GDataReader* reader);


  /** Destroys the data exchange object. */
  ~GDataExchange(void);


  /** Finds a free port in the range <code>GDE_PORTRANGE_LO</code> and 
      <code>GDE_PORTRANGE_HI</code> and creates a server socket listening on this
      port. Raises a <code>connectionStateSignal</code> with information
      about the status of the connection, and also prints the status information
      to <code>stdout</code>.
      @return <code>true</code> if the server socket was set up successfully,
        <code>false</code> otherwise. */
  bool setupServerSocket(void);


  /** Asynchronously sends out data.
      @param data the data to be sent.
      @param size the size of the data. */
  void asyncSend(unsigned char* data, unsigned int size);


  /** Synchronously sends out data. Does not return before the data has
      been sent. This method performs the same functionality as
      <code>write(...)</code> from the C library.
      @param data the data to be sent
      @param size the size of the data
      @return the number of bytes written, or <code>-1 (0xFFFFFFFF)</code> in case of
        an error (<code>errno</code> can be queried to get additional
	information since this function directly calls <code>write(...)</code>). */
  unsigned int syncSend(unsigned char* data, unsigned int size);


  /** Synchronously reads data. If no data is available, this method immediately returns
      (non-blocking I/O).
      @param buffer the buffer the data is to be stored in
      @param size the size of the buffer.
      @return the number of bytes read, or <code>-1 (0xFFFFFFFF)</code> if no data
        is available or in case of an error (<code>errno</code> can be queried to get
        additional information since this function directly calls <code>read(...)</code>). */
  unsigned int syncRead(unsigned char* buffer, unsigned int size);


 signals:
  /** Raised when the status of the socket connection has changed.
      @param status the new connection status (a <code>GDE_STATUS_xxxx</code> constant)
      @param hostname the name of the host Grisu is running on (may be <code>NULL</code>)
      @param port the numer of the port used in the connection (may be <code>-1</code>) */
  void connectionStateSignal(unsigned char status, char* hostname, int port);


 private slots:
  /** Called when the server socket accepted a connection.
      @param sock the socket that was created after accepting the connection. */
  void acceptedSlot(KSocket* sock);

 
 /** Called when data is available for reading. After all available data was read, 
     it is passed on to <code>dataReader->processData</code>.
     @param sock not used. */
  void readSlot(KSocket* sock);


  /** Called when the socket is ready for writing. Writes all data on <code>outBuffer</code>
      to the socket, then deletes the buffer.
      @param sock not used. */
  void writeSlot(KSocket* sock);


  /** Called when the socket connection was closed.
      @param sock not used. */
  void closeSlot(KSocket* sock);
};


#endif
