/*  $Id: glatexrenderer.cpp,v 1.8 2003/05/22 16:43:11 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Implementation of the <code>GLatexRenderer</code> class.

    @author Holger Wunsch */


#include "config.h"
#include "glatexrenderer.h"


/** Creates a LaTeX renderer.
    @param a_data the data
    @param a_drawMode either <code>DRAW_STRUC</code>, <code>DRAW_TREE</code>, or
    <code>DRAW_AUTO</code>. If <code>DRAW_AUTO</code> is given, tree output is generated
    when tree data is available, otherwise the data is rendered as feature structure.
    @param a_dispNodes the displayable node data
    @param a_outName the name of the output file. */
GLatexRenderer::GLatexRenderer(Document* a_data, int a_drawMode, IdTree* a_dispNodes, const QString& a_outName) {
  data = a_data;

  if (a_drawMode == DRAW_AUTO) {
    if (data->tree != NULL)
      drawMode = DRAW_TREE;
    else
      drawMode = DRAW_STRUC;
  }
  else
    drawMode = a_drawMode;

  dispNodes = a_dispNodes;
  outName = a_outName;
}


/** Destroys this instance. */
GLatexRenderer::~GLatexRenderer(void) {
}


/** Renders the node data.
    @return <code>true</code> if the rendering process was successful,
    <code>false</code> if an error occurred. */
bool GLatexRenderer::render(void) {
  outFile = fopen(outName, "w");
  if (outFile == NULL)
    return false;

  if (writeLatexHeader() == false) {
    fclose(outFile);
    return false;
  }
  
  if (drawMode == DRAW_STRUC) {
    writeAVMHeader();
    
    if (renderGenericNode(data->data->getFirst(), NodeAddress()) == false) {
      fclose(outFile);
      return false;
    }
    
    writeAVMFooter();
  }
  else {
    if (renderTreeNode(data->tree, NodeAddress()) == false) {
      fclose(outFile);
      return false;
    }
  }

  if (writeLatexFooter() == false) {
    fclose(outFile);
    return false;
  }
  
  fclose(outFile);
  
  return true;
}


/** Writes the LaTeX header to the output file.
    @return <code>true</code> if the header was written successfully,
    <code>false</code> otherwise. */
bool GLatexRenderer::writeLatexHeader(void) {
  fprintf(outFile, "%% AVM output generated by Grisu%s\n\n", GRISU_VERSION);
  fputs("\\documentclass{article}\n", outFile);
  fputs("\\usepackage{avm+}\n", outFile);
  
  if (drawMode == DRAW_TREE)
    fputs("\\usepackage{ecltree+}\n", outFile);

  fputs("\\avmoptions{center}\n", outFile);

  fputs("\\begin{document}\n", outFile);

	  
  return true;
}


/** Writes the necessary LaTeX closing statements to the output file. 
    @return <code>true</code> if the header was written successfully,
    <code>false</code> otherwise. */
bool GLatexRenderer::writeLatexFooter(void) {

  fputs("\\end{document}\n", outFile);

  return true;
}


/** Writes the opening statements of an <code>avm</code> environment.
    @return <code>true</code> if the statement was written sucessfully, <code>false</code>
    otherwise. */
bool GLatexRenderer::writeAVMHeader(void) {
  fputs("\\begin{avm}\n", outFile);

  return true;
}


/** Writes the closing statements of an <code>avm</code> environment.
    @return <code>true</code> if the statement was written sucessfully, <code>false</code>
    otherwise. */
bool GLatexRenderer::writeAVMFooter(void) {
  fputs("\\end{avm}\n", outFile);

  return true;
}


/** Renders a generic node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderGenericNode(GenericNode* a_node, const NodeAddress& a_addrParent) {
  switch (a_node->nodetype) {
  case NODETYPE_ATOM_NODE:
    return renderAtomNode((AtomNode*) a_node, a_addrParent);
  case NODETYPE_CONJUNCTION_NODE:
    return renderConjunctionNode((ConjunctionNode*) a_node, a_addrParent);
  case NODETYPE_DISJUNCTION_NODE:
    return renderDisjunctionNode((DisjunctionNode*) a_node, a_addrParent);
  case NODETYPE_FEATVAL_NODE:
    return renderFeatvalNode((FeatvalNode*) a_node, a_addrParent);
  case NODETYPE_LIST_NODE:
    return renderListNode((ListNode*) a_node, a_addrParent);
  case NODETYPE_REENT_NODE:
    return renderReentNode((ReentNode*) a_node, a_addrParent);
  case NODETYPE_REF_NODE:
    return renderRefNode((RefNode*) a_node, a_addrParent);
  case NODETYPE_SET_NODE:
    return renderSetNode((SetNode*) a_node, a_addrParent);
  case NODETYPE_STRUC_NODE:
    return renderStrucNode((StrucNode*) a_node, a_addrParent);
  case NODETYPE_TAIL_NODE:
    return renderTailNode((TailNode*) a_node, a_addrParent);
  case NODETYPE_TREE_NODE:    
    return renderTreeNode((TreeNode*) a_node, a_addrParent);
  case NODETYPE_TYPE_NODE:
    return renderTypeNode((TypeNode*) a_node, a_addrParent);
  }
   
  return true;
}

  
/** Renders an atom node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderAtomNode(AtomNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);
  
  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;
  
  if ((flags & NODEFLAG_HIDDEN) == 0) {
    fputs(latexifyString(a_node->atom), outFile);
    return true;
  }
  else
    return false;
}


/** Renders a conjunction node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderConjunctionNode(ConjunctionNode* a_node, const NodeAddress& a_addrParent) {
  printf("GLatexRenderer::renderConjunctionNode not implemented\n");
  
  return false;
}


/** Renders a disjunction node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderDisjunctionNode(DisjunctionNode* a_node, const NodeAddress& a_addrParent) {
  printf("GLatexRenderer::renderDisjunctionNode not implemented\n");

  return false;
}
  

/** Renders a featval node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderFeatvalNode(FeatvalNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);
  
  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if ((flags & NODEFLAG_HIDDEN) == 0) {
    fprintf(outFile, "%s & ", latexifyString(a_node->feature));
    
    renderGenericNode(a_node->value, l_addr);

    return true;
  }
  else
    return false;
}


/** Renders a function node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderFunctionNode(FunctionNode* a_node, const NodeAddress& a_addrParent) {
  printf("GLatexRenderer::renderFunctionNode not implemented\n");

  return false;
}


/** Renders a list node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderListNode(ListNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if ((flags & NODEFLAG_HIDDEN) == 0) {
    List* structs = a_node->structs;
    bool drawn;

    fputs("\\<\n", outFile);
    
    while (structs) {
      drawn = renderGenericNode(structs->item, l_addr);
      structs = structs->tail;
      if (structs && drawn)
	fputs("$\\,$,\n", outFile);
    }
    
    if (a_node->tail) {
      fputs("$\\mid$\n", outFile);
      renderTailNode(a_node->tail, l_addr);
    }
    
    fputs("\\>\n", outFile);

    return true;
  }
  else
    return false;
}


/** Renders a reent node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderReentNode(ReentNode* a_node, const NodeAddress& a_addrParent) {
  return renderGenericNode(a_node->value, a_addrParent);
}


/** Renders a ref node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderRefNode(RefNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if ((flags & NODEFLAG_HIDDEN) == 0) {
    fprintf(outFile, "\\@{%d} ", a_node->ref);
    
    if (flags & NODEFLAG_EXPANDED) {
      ReentNode* reentNode = data->findReentrancy(a_node->ref);
      if (reentNode != NULL)
	renderReentNode(reentNode, l_addr);
    }

    return true;
  }
  else
    return false;
}


/** Renders a rest node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderRestNode(RestNode* a_node, const NodeAddress& a_addrParent) {
  return renderGenericNode(a_node->rest, a_addrParent);
}


/** Renders a set node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderSetNode(SetNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if ((flags & NODEFLAG_HIDDEN) == 0) {
    List* structs = a_node->structs;
    bool drawn;
    
    fputs("\\{\n", outFile);
    
    while (structs) {
      drawn = renderGenericNode(structs->item, l_addr);
      structs = structs->tail;
      if (structs && drawn)
	fputs("$\\,$,\n", outFile);
    }
    
    if (a_node->rest) {
      fputs("$\\mid$\n", outFile);
      renderRestNode(a_node->rest, l_addr);
    }
    
    fputs("\\}\n", outFile);

    return true;
  }
  else
    return false;
}



/** Renders a struc node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderStrucNode(StrucNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;
  
  if ((flags & NODEFLAG_HIDDEN) == 0) {
    List* featvals = a_node->featvals;
    bool drawn;
    
    if (featvals) {
      fputs("\\[\\tp{", outFile);
      renderTypeNode(a_node->type, l_addr);
      fputs("}\\\\\n", outFile);
      
      while (featvals) {
	drawn = renderGenericNode(featvals->item, l_addr);
	
	featvals = featvals->tail;
	if (featvals != NULL && drawn)
	  fputs("\\\\\n", outFile);
      }
      
      fputs("\\]\n", outFile);
    }
    else {
      //      fputs("\\tp{", outFile);
      renderTypeNode(a_node->type, l_addr);
      fputs("\n", outFile);
      //      fputs("}\n", outFile);
    }

    return true;
  }
  else
    return false;
}


/** Renders a tail node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderTailNode(TailNode* a_node, const NodeAddress& a_addrParent) {
  return renderGenericNode(a_node->tail, a_addrParent);
}


/** Renders a tree node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderTreeNode(TreeNode* a_node, const NodeAddress& a_addrParent) {
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  unsigned long flags;
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if (a_addrParent.len > 0) {
    // for inner nodes, always write out \chunk
    fputs("\\chunk", outFile);
    if (a_node->arclabel != NULL)
      fprintf(outFile, "[%s]", a_node->arclabel);
    fputs("{\n", outFile);
  }

  if (a_node->subtrees != NULL || a_addrParent.len == 0)
    fputs("\\begin{bundle}{", outFile);

  if (flags & NODEFLAG_EXPANDED) {
    fprintf(outFile, "\\begin{Avm}{%s}\n", a_node->label);
    GenericNode* struc = data->findNodeById(a_node->linkid);
    if (struc)
      renderGenericNode(struc, l_addr);
    fputs("\\end{Avm}\n", outFile);
  }
  else
    fprintf(outFile, "%s\n", a_node->label);

  if (a_node->subtrees != NULL || a_addrParent.len == 0)
    fputs("}\n", outFile);

  List* subtrees = a_node->subtrees;
  while (subtrees != NULL) {
    renderGenericNode(subtrees->item, l_addr);
    subtrees = subtrees->tail;
  }

  if (a_node->subtrees != NULL || a_addrParent.len == 0)
    fputs("\\end{bundle}\n", outFile);

  if (a_addrParent.len > 0)
    fputs("}\n", outFile); // end of \chunk

  return true;
}


/** Renders a type node.
    @param a_node the node.
    @param a_addrParent the node address of the parent node.
    @return <code>true</code> if the node was written, <code>false</code> if not,
    for example because the node is hidden. */
bool GLatexRenderer::renderTypeNode(TypeNode* a_node, const NodeAddress& a_addrParent) {
  unsigned long flags;
  NodeAddress l_addr(a_addrParent, (GenericNode*) a_node);
  DisplayableNode* l_dn = dispNodes->findDisplayableNode(l_addr);

  if (l_dn != NULL)
    flags = l_dn->flags;
  else
    flags = a_node->flags;

  if ((flags & NODEFLAG_HIDDEN) == 0) {
    fprintf(outFile, "%s", latexifyString(a_node->name));

    return true;
  }
  else
    return false;
}


/** Modifies a string such that it prints properly in LaTeX.
    
    In LaTeX, several characters are command characters which must be
    escaped to make them print properly. These charaters are:

    <pre>$, %, {, }, &, #, _, ^</pre>

    @param a_inStr the string to be modified
    @return a pointer to a static buffer which contains the modified string. */
char* GLatexRenderer::latexifyString(const char* a_inStr) {
  static char l_outStr[LATEXIFY_BUFFER_SIZE];
  char* l_outPtr = l_outStr;
  char* l_end = &(l_outStr[LATEXIFY_BUFFER_SIZE - 2]);
  const char* l_inPtr = a_inStr;

  while (*l_inPtr != '\x0') {
    if (*l_inPtr == '$' || 
	*l_inPtr == '%' || 
	*l_inPtr == '{' ||
	*l_inPtr == '}' ||
	*l_inPtr == '&' ||
	*l_inPtr == '#' ||
	*l_inPtr == '_' ||
	*l_inPtr == '^') {
      *(l_outPtr++) = '\\';
      if (l_outPtr == l_end) {
	printf("GLatexRenderer::latexifyString: Buffer size exceeded.\n");
	break;
      }
	
      *(l_outPtr++) = *l_inPtr;
      if (l_outPtr == l_end) {
	printf("GLatexRenderer::latexifyString: Buffer size exceeded.\n");
	break;
      }
    }
    else {
      *(l_outPtr++) = *l_inPtr;
      if (l_outPtr == l_end) {
	printf("GLatexRenderer::latexifyString: Buffer size exceeded.\n");
	break;
      }
    }      

    l_inPtr++;
  }

  *l_outPtr = '\x0';

  return l_outStr;
}
