/*  $Id: idtree.cpp,v 1.29 2003/05/22 16:43:12 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Implementation of the class IdTree.

    @author Holger Wunsch */


#include <string.h>
#include <stdlib.h>
#include "idtree.h"


/** Creates a new, empty map that can hold up to <code>max</code> mappings.
    @param max the maximum number of mappings. */
IdTree::IdTree(unsigned int max) {
  maxMappings = max;
  addrs = new NodeAddrMappingPtr[maxMappings];
  memset(addrs, 0, sizeof(NodeAddrMappingPtr) * maxMappings);
}
  

/** Destroys the entire map. */
IdTree::~IdTree(void) {
  for (unsigned int i = 0; i < maxMappings; i++) {
    if (addrs[i] != NULL) {
      NodeAddrMapping* m = addrs[i];
      NodeAddrMapping* mdel;
      while (m != NULL) {
	mdel = m;
	m = m->next;
	delete mdel;
      }
    }
  }

  delete [] addrs;
}


/** Adds a node to the map. If the node has been added previously,
    its corresponding DisplayableNode is returned. 
    @param na the node address of the node
    @param node a pointer to the node data
    @return a DisplayableNode representing the visible instance
    of the node. */
DisplayableNode* IdTree::addDisplayableNode(const NodeAddress& na, GenericNode* node) {
  DisplayableNode* dn;
  unsigned int id = atoi(node->id);

  if (addrs[id] == NULL) {
    // No mappings in list so far
    NodeAddrMapping* m = new NodeAddrMapping;
    dn  = new DisplayableNode;
    dn->addr = na;
    dn->node = node;
    dn->flags = node->flags;
    m->node = dn;
    m->next = NULL;
    addrs[id] = m;
  }
  else {
    // There are some mappings in the list
    NodeAddrMapping* m = addrs[id];
    NodeAddrMapping* mprev;
    while (m != NULL) {
      if (m->node->addr == na) {
	// Node address already in list, just return it
	dn = m->node;
	break;
      }
      mprev = m;
      m = m->next;
    };
    if (m == NULL) {
      // Add a new mapping
      m = new NodeAddrMapping;
      dn  = new DisplayableNode;
      dn->addr = na;
      dn->node = node;
      dn->flags = node->flags;
      m->node = dn;
      m->next = NULL;
      mprev->next = m;
    }
  }

  return dn;
}
    
      
/** Looks up a DisplayableNode in the map given the node address of
    the corresponding node.
    @param na the node address
    @return The DisplayableNode or <code>NULL</code> if the no node wirh the
    given address exists in the id tree. */
DisplayableNode* IdTree::findDisplayableNode(const NodeAddress& na) {
  unsigned int id = atoi(na.nodeID());
  DisplayableNode* dn = NULL;

  if (addrs[id] != NULL) {
    NodeAddrMapping* m = addrs[id];
    while (m) {
      if (na == m->node->addr) {
	dn = m->node;
	break;
      }
      m = m->next;
    }
  }

  return dn;
}
