/*  $Id: options.h,v 1.29 2003/05/22 16:43:14 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file
    
    Management of user-definable application options

    @author Holger Wunsch */


#ifndef __OPTIONS_H_
#define __OPTIONS_H_

#include <qwidget.h>
#include <qstrlist.h>
#include <qprinter.h>


/** Management of application options.
    @author Holger Wunsch */
class Options:public QObject {
  Q_OBJECT

public:
  /** Contants in this enum are passed to changeColorSlot in order to specify
      what color value is to be changed. */
  enum COLORTYPES {
    BackgroundColor = 1,
    NormalColor     = 2,
    TypeColor       = 3,
    TextColor       = 4,
    LeafColor       = 5,
    TreeColor       = 6,
    TreeLineColor   = 7,
    ErrorColor      = 8,
    HiddenColor     = 9,
    MarkerColor     = 10,
    FoundColor      = 11,
    DifferentNodeColor = 12,
    ArcLabelColor   = 13,
    ArcLabelBgColor   = 14
  };

  /** Contants in this enum are passed to changeFontSlot in order to specify
      which font is to be changed. */
  enum FONTTYPES {
    NormalFont      = 1,
    TypeFont        = 2,
    TextFont        = 3,
    LeafFont        = 4,
    TreeFont        = 5,
    ErrorFont       = 6,
    ArcLabelFont    = 7
  };


public:
  /** The background color. */
  QColor backgroundColor;


  /** The color of features, tags and other visual objects like brackets. */
  QColor normalColor;


  /** The color of types. */
  QColor typeColor;


  /** The color of atoms. */
  QColor textColor;


  /** Error color. */
  QColor errorColor;


  /** The color of inner nodes in a tree. */
  QColor treeColor;


  /** The color of the lines connecting nodes in a tree. */
  QColor treeLineColor;


  /** The color of leaves in a tree. */
  QColor leafColor;


  /** The color of hidden nodes. */
  QColor hiddenColor;


  /** The color of a structure marker (used in highlight). */
  QColor markerColor;


  /** The color of a found marker. */
  QColor foundColor;


  /** The color of a marker indicating different nodes (in a diff). */
  QColor differentNodeColor;


  /** The text color the text of an arc label in a tree. */
  QColor arclabelColor;


  /** The background color of an arc label in a tree. */ 
  QColor arclabelBgColor;
  

  /** The font of features, tags and other visual objects like brackets. */
  QFont  normalFont;


  /** The font of types. */
  QFont  typeFont;

  
  /** The font of atoms. */
  QFont  textFont;


  /** The font of inner tree nodes. */
  QFont  treeFont;


  /** The font of leaves in a tree. */
  QFont  leafFont;


  /** Error font. */
  QFont  errorFont;


  /** The font of edge labels in a tree. */
  QFont  arclabelFont;
  

  /** The width of the lines representing edges in a tree. */
  int treeLineWidth;


  /** The amount of space between nodes (the "Structure Spacing" value modifiable
      in the GUI). */
  int offset;


  /** <code>true</code> if grayed-out placeholders should be shown for hidden nodes. */
  bool drawHidden;


  /** <code>true</code> if view windows should automatically adjust their size to that
      of the structures displayed. */
  bool fitToDrawing;


  /** <code>true</code> if a view window is to be opened automatically when new data
      arrives. */
  bool autoOpenWindows;


  /** <code>true</code> if reentrancies are to expanded automatically when a window
      opens. */
  bool expandReentrancies;


  /** <code>true</code> if structures are to be highlighted when the user points on them
      with the mouse. */
  bool highlightStructures;

  
  /** Currently not used. */
  bool writeConfig;

  
  /** If this option is set to <code>true</code>, the application name
      is shown in view windows in addition to the name supplied by the
      document. */
  bool showAppName;


  /** If this option is set to <code>true</code>, KFileDialog is used
      instead of QFileDialog in all file operations. */
  bool useKFileDialog;


public:
  /** Creates a new Options object with default values. */
  Options();


  /** Copy constructor. */
  Options(Options &);


  /** Destroys the options instance. */
  ~Options();


  /** Writes all settings in the given printer dialog to the configuration file.
      @param printer the <code>QPrinter</code> object. */
  void writePrinterOptions(QPrinter& printer);


  /** Initializes a printer dialog with options from the configuration file.
      @param printer the <code>QPrinter</code> object. */
  void readPrinterOptions(QPrinter& printer);
      

public slots:
  /** Shows the color dialog
  
      @param what Specifies the color value to be changed (one of the COLORTYPES constants).
      @param parent the parent widget. */
  void changeColorSlot(int what, QWidget* parent);
 
 
  /** Shows the font dialog.

      @param what Specifies what font is to be changed (one of the FONTTYPES constants). */
  void changeFontSlot(int what);


  /** Reads in all options from a KDE-style config file (Grisurc). */
  void readOptionFromConfig();


  /** Writes all options to a KDE-style config file (Grisurc). */
  void writeOptionFromConfig();


  /** Copies all options from <code>op</code> into <code>this</code>. */
  void copyFromLocal(Options &);

};

#endif
