/*  $Id: vcanvas.h,v 1.34 2003/05/24 01:55:03 dm Exp $

    This file is part of the

    GRammar vISUalization tool (GRISU)

    Developed by Holger Wunsch and Detmar Meurers
    Copyright (C) 1998-2003 by the developers
    All rights reserved

    Seminar fr Sprachwissenschaft, Universitt Tbingen and
    Department of Linguistics, The Ohio State University

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file

    Declarations for class ViewCanvas.

    @author Holger Wunsch */


#ifndef __VCANVAS_H_
#define __VCANVAS_H_
#include <qdir.h>



#include <qscrbar.h>
#include <qwidget.h>
#include <qkeycode.h>

#include <kapp.h>
#include <kmainwindow.h>
#include <kiconloader.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <kmenubar.h>
#include <kstatusbar.h>

#include "data.h"
#include "canvas.h"
#include "scrollcanvas.h"
#include "NodeAddress.h"

#include "StrucFindDialog.h"


/* Menuitem IDs */
#define MID_VCOPTIONS_FIT           301  /**< Menuitem ID */
#define MID_VCOPTIONS_HIGHLIGHT     302  /**< Menuitem ID */
#define MID_VCOPTIONS_SHOWHIDDEN    303  /**< Menuitem ID */


/** The ViewCanvas displays the window containing feature structures 
    and trees.

    The ViewCanvas class is one of the two central user interface
    classes in Grisu. The secondary windows that display feature
    structures or trees are instances of ViewCanvas. All user
    interactions with menus and toolbars are handled in this class.

    @author Holger Wunsch */
class ViewCanvas:public KMainWindow {
  Q_OBJECT


  /** The document displayed in this ViewCanvas. */
  Document *document;

  
  /** An instance of the Options class that specifies the user-set
      visible properties. */
  Options *options;


  /** An instance of a Finder instance, if the user issued a
      "find structure" command. Otherwise, this is <code>NULL</code>. */
  Finder* finder;


  /** The main menu. */
  KMenuBar *menuBar;


  /** The file menu */
  QPopupMenu *fileMenu;


  /** The edit menu. */
  QPopupMenu *editMenu;


  /** The data menu. */
  QPopupMenu* dataMenu;


  /** The options menu. */
  QPopupMenu *optionsMenu;


  /** The window menu. */
  QPopupMenu *windowMenu;
  

  /** The main toolbar. */
  KToolBar *mainToolBar;
  

  /** If a structure find dialog was opened by the user, this points
      to the instance of the dialog, otherwise <code>NULL</code>. */
  StrucFindDialog* findDlg;


  /** Saves the current structure or tree after prompting the
      user for the file name. */
  void saveStructure(void);


protected:
  /** Closes this window and emits a <code>closing</code> signal.
      
      @param ce the event class */
  void closeEvent(QCloseEvent* ce);


public:
  /** A pointer to the Canvas that does the actual drawing. */
  Canvas *canvas;

  
  /** Takes care of scrollbars. */
  ScrollCanvas* scrollCanvas;


  /** The current display mode of the data (structure or tree). */
  int showing;

 
  /** The status bar. */
  KStatusBar* statusBar;


  /** The constructor sets up the view window. It initializes the menus and toolbars and
      creates a ScrollCanvas which in turn contains a Canvas, which takes care of the
      actual displaying of structures and trees.

      @param d the document (data package) to be shown in this window.
      @param what specifies whether the data is to presented as structure (<code>DRAW_STRUC</code>),
      or as tree (<code>DRAW_TREE</code>).
      @param op a pointer to the global options object in MainWindow (MainWindow::globalOptions)
      @param parent the parent widget
      @param name this widget's name */
  ViewCanvas(Document* d, int what, Options* op, QWidget *parent=0, const char *name=0);


  /** Destroys this window, and all dependent windows (find dialogs, for example). */
  ~ViewCanvas();


  /** Called in reponse to a resize event.
      
  @param ev the event class. */
  virtual void resizeEvent(QResizeEvent* ev);
  

  /** Redraws this window. Does nothing but emit a <code>redrawSignal</code>.*/
  void redraw();


  /** Expands the first occurrence of any reentrancy within a structure.

      @param node The node data of the node to be expanded.
      @param addr The node address.
      @param expRef A list of IDs of that have been expanded so far. If later in
      the structure, a node with an ID is encountered that is already in this list,
      this node won't be expanded any more. */
  void expandReentrancies(GenericNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a TreeNode.

      @param node the TreeNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesTreeNode(TreeNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a StrucNode.

      @param node the StrucNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesStrucNode(StrucNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a FeatvalNode.

      @param node the FeatvalNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesFeatvalNode(FeatvalNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a ListNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesListNode(ListNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a TailNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesTailNode(TailNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a SetNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesSetNode(SetNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a RestNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesRestNode(RestNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands a RefNode. If the node ID of this RefNode isn't in the <code>expRef</code> list yet,
      the <code>NODEFLAG_EXPANDED</code> flag is set on this node's <code>DisplayableNode</code>.
      
      @param node the RefNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesRefNode(RefNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a FunctionNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesFunctionNode(FunctionNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a RelationNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesRelationNode(RelationNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a DisjunctionNode.

      @param node the DisjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesDisjunctionNode(DisjunctionNode* node, NodeAddress addr, QList<int>* expRef);


  /** Expands all RefNodes that occur in operands of a ConjunctionNode.

      @param node the ConjunctionNode
      @param addr the unique node address of the displayed node
      @param expRef a list of linkids of node that have already been expanded to avoid multiple expansion. */
  void expandReentranciesConjunctionNode(ConjunctionNode* node, NodeAddress addr, QList<int>* expRef);


  /** Tests whether this windows displays the given document.

      @param doc the document in question
      @return <code>true</code> if this window displays the given document, 
      <code>false</code> otherwise. */
  bool showsDoc(Document* doc);


  /** Called when the user issued a search command from a context menu. This will start
      a search beginning at the node the user clicked on.
      @param searchTerm the search term (usually the attribute's or a type's name etc.)
      @param addr the address of the node the user clicked on. */
  void contextMenuSearch(char* searchTerm, const NodeAddress addr);


  /** Changes the status bar message.
      @param message the new message. */
  void setStatusBarMessage(const char* message);


signals:

  /** This signal is emitted when the window must be repainted. */
  void redrawSignal();


  /** This signal is raised when the previous available structure is to be shown
      on user request. 

      @param replace must be set to <code>this</code>. */
  void previousStructureSignal(ViewCanvas* replace);


  /** This signal is raised when the next available structure is to be shown
      on user request. 

      @param replace must be set to <code>this</code>. */
  void nextStructureSignal(ViewCanvas* replace);


  /** Emitted when this window closes. This is used to notify Grisu's main window, which
      maintains a list of all open ViewCanvas windows, that this window was closed.

      @param thisWindow a pointer to <code>this</code>. */
  void closing(ViewCanvas* thisWindow);


public slots:
  /** Called when the user clicked on the File|Save menu item. */
  void saveSlot();


  /** Repaints the contents of the Canvas object, then recalulates the
      window size (if the 'window fits structure size' option is turned on). */
  void redrawSlot();


  /** Toggles the 'window fits structure size' option, then repaints the window. */
  void toggleFitSlot(void);


  /** Toggles the 'show hidden nodes' option, then repaints the window. */
  void toggleShowHiddenSlot(void);


  /** Toggles the 'highlight structures' option, then repaints the window. */
  void toggleHighlightSlot(void);


  /** Calls <code>expandReentrancies</code>, then repaints the window. */
  void expandReentranciesSlot();


  /** Closes the window. */
  void quitWindowSlot();


  /** Switches the display mode to show a tree (if tree data is available. If there is no
      tree data, a structure is displayed instead). */
  void showTreeSlot();


  /** Switches the display mode to show a structure */
  void showStructureSlot();


  /** Opens a new ViewCanvas window, then shows the document displayed in the current
      window in the new window as a structure. */
  void showStructureNewWindowSlot();


  /** Opens a new ViewCanvas window, then shows the document displayed in the current
      window in the new window as a tree (if tree data is available. If there is no
      tree data, a structure is displayed instead).*/
  void showTreeNewWindowSlot();


  /** Prints the contents of this window. */
  void printSlot();


  /** Opens a find dialog. */
  void searchSlot();


  /** Starts a new search with the argument entered in the find dialog. */
  void findSlot();


  /** Finds the next occurence of a search term. */
  void findNextSlot();


  /** Closes the find dialog and deletes the current <code>Finder</code> isnatnce. */
  void cancelSearchSlot();


  /** Resets the structure so that it is displayed in the way it was sent by the backend
      (with respect to expanded nodes, hidden nodes etc.) */
  void resetStructureSlot();


  /** Called in response to a user's hide/unhide node request in a popup menu. */
  void hideUnhidePopupSlot();


  /** Called in response to a user's expand reentrancy request in a popup menu. */
  void expandLocalPopupSlot();


  /** Shows the options dialog. */
  void optionsSlot();


  /** Raises the main window (shows the main window on top of all other windows). */
  void raiseMainWindowSlot(void);


  /** Shows the previous document in the list of available documents. If tree data
      is available, the document is displayed as a tree, otherwise as a structure
      (unless the document explictly overrides this by specifying <code>!newdata(tree)</code>
      or <code>!newdata(structure)</code>.<br><br>
      This will actually close the current window and open a new window with the same
      size and position.*/
  void previousStructureSlot(void);


  /** Shows the previous document in the list of available documents. If tree data
      is available, the document is displayed as a tree, otherwise as a structure
      (unless the document explictly overrides this by specifying <code>!newdata(tree)</code>
      or <code>!newdata(structure)</code>.<br><br>
      This will actually close the current window and open a new window with the same
      size and position.*/
  void nextStructureSlot(void);


  /** Writes the currently selected data package to a LaTeX file. */
  void renderLatexSlot(void);
};

#endif
